// ----------------------------------------------------------------------------
//      data_io.cxx
//
// Copyright (C) 2014
//              Stelios Bounanos, M0GLD
//              David Freese, W1HKJ
//
// This file is part of fldigi.
//
// fldigi is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or
// (at your option) any later version.
//
// fldigi is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>
#include <strstream>

#include <FL/Fl.H>
#include <FL/fl_ask.H>

#include "data_io.h"
#include "util.h"
#include "flaa.h"
#include "flaa_ui.h"
#include "status.h"
#include "serial.h"

guard_lock::guard_lock(pthread_mutex_t* m, int h) : mutex(m) {
	pthread_mutex_lock(mutex);
}

guard_lock::~guard_lock(void) {
	pthread_mutex_unlock(mutex);
}

static pthread_t read_thread;
static pthread_mutex_t read_mutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_mutex_t status_mutex = PTHREAD_MUTEX_INITIALIZER;

bool read_exit = false;
bool rx_data = false;
bool cable_values = false;
bool thread_running = false;

static void read_data();
static void read_cable_values();
static void read_swr(int band);
static bool get_rx(int F, double &R, double &X);

#define LOOPTIME 50

void * readloop(void *)
{
	while (1) {
		MilliSleep(LOOPTIME);
		if (read_exit) break;
		{
			guard_lock readlock( &read_mutex );
			if (rx_data) {
				read_data();
				rx_data = false;
			}
			if (flaa_status.swr_80_on) read_swr(80);
			if (flaa_status.swr_60_on) read_swr(60);
			if (flaa_status.swr_40_on) read_swr(40);
			if (flaa_status.swr_30_on) read_swr(30);
			if (flaa_status.swr_20_on) read_swr(20);
			if (flaa_status.swr_15_on) read_swr(15);
			if (flaa_status.swr_10_on) read_swr(10);
			if (flaa_status.swr_6_on) read_swr(5);
			if (flaa_status.swr_2_on) read_swr(2);
			if (flaa_status.swr_440_on) read_swr(440);

			if (cable_values) {
				read_cable_values();
				cable_values = false;
			}
		}
	}
	return NULL;
}

bool start_read_thread()
{
	if (thread_running) return false;  // ONLY ONE instance allowed

	read_exit = false;

	if (pthread_create(&read_thread, NULL, readloop, NULL) < 0) {
		fl_alert("arq init: pthread_create failed");
		thread_running = false;
		return false;
	}

	thread_running = true;
	return true;
}

void exit_read_thread() {
	if (thread_running == false) return;

	read_exit = true;

	pthread_join(read_thread, NULL);
	thread_running = false;
}

static string status_txt;
static void update_status(void *)
{
	if (btn_toggle_status->value() == 0) return;

	guard_lock status_lock(&status_mutex);
	if (status_txt.empty()) return;
#ifdef DEBUG
std::cout << "Status: " << hexstr(status_txt);
#endif
	size_t p = status_txt.find("\n");
	if (p != string::npos) status_txt.erase(p);
	txt_status->value(status_txt.c_str());
	txt_status->redraw();
	status_txt.clear();
}

bool write_command(string cmd)
{
	int numtries = 0;
	int nread = 0;
	string data;
	aa_ser_port.WriteBuffer(cmd.c_str(), cmd.length());
	status_txt.assign("Write: ").append(cmd);
	Fl::awake(update_status);
//std::cout << "Write: " << hexstr(cmd);

	numtries = 100;
	memset(serbuff, 0, BUFFSIZE);
	while ((nread = aa_ser_port.ReadBuffer(serbuff, 128)) == 0) {
		MilliSleep(10);
		numtries--;
		if (!numtries) return false;
		if (read_exit) return false;
	}
	data.assign(serbuff);
	if (!numtries || (data.find("OK") == string::npos))
		return false;
	return true;
}

void read_data()
{
	swr_canvas->clear();
	rx_canvas->clear();

	string data;
	string linebuff;
	int nread = 0;
	int numtries = 100;
	size_t p = 0;

	double midf = (flaa_status.fmax + flaa_status.fmin) / 2.0;
	double delf = (flaa_status.fmax - flaa_status.fmin);
	double imidf = midf * 1e6;
	int idelf = delf * 1e6;
	int numread = 0;

	char szcmd[50];

	if (!write_command("ON\n")) goto exit_read;

	imidf *= (1 + flaa_status.freq_corr / 1e6);
	snprintf(szcmd, sizeof(szcmd), "FQ%d\n", (int)imidf);
	if (!write_command(szcmd)) goto exit_read;

	snprintf(szcmd, sizeof(szcmd), "SW%d\n", idelf);
	if (!write_command(szcmd)) goto exit_read;

	snprintf(szcmd, sizeof(szcmd), "FRX%d\n", flaa_status.npts);
	aa_ser_port.WriteBuffer(szcmd, strlen(szcmd));

	status_txt.assign("Write: ").append(szcmd);
	Fl::awake(update_status);
//std::cout << "Write: " << hexstr(szcmd);

	numtries = 0;
	data.clear();
	
	char szstatus[64];
	memset(serbuff, 0, BUFFSIZE);
	for (;;) {
		if (read_exit) goto exit_read;
		memset(serbuff, 0, BUFFSIZE);
		nread = aa_ser_port.ReadBuffer(serbuff, BUFFSIZE);
		if (nread) {
			numread += nread;
			data.append(serbuff);
			snprintf(szstatus, sizeof(szstatus), "Read: %4d bytes\n", numread);
			{
				guard_lock status_lock(&status_mutex);
				status_txt = szstatus;
				Fl::awake(update_status);
			}
		}

		if (data.find("OK") != string::npos)  break;
		if (data.find("ERROR") != string::npos)  goto exit_read;

		MilliSleep(10);
		numtries++;
		if (numtries * 0.01 > 30.0) {
			guard_lock status_lock(&status_mutex);
			status_txt = "READ FAILURE\n";
			Fl::awake(update_status);
			 goto exit_read;
		}
	}
	{
		guard_lock status_lock(&status_mutex);
		status_txt = "Scan completed\n";
		Fl::awake(update_status);
	}

	while ((p = data.find("\r")) != string::npos) data.erase(p,1);
	p = data.find("\nOK\n");
	if (p != string::npos) data.erase(p);

	float Fr, Rr, Xr;
	testdata.clear();
	p = data.find("\n");
	while (p != string::npos) {
		linebuff = data.substr(0, p);
		sscanf(linebuff.c_str(), "%f,%f,%f", &Fr, &Rr, &Xr);
		Fr /= (1 + flaa_status.freq_corr / 1e6);
		testdata.push_back(FRX(Fr, Rr, Xr));
		data.erase(0, p+1);
		p = data.find("\n");
	}
//std::cout << "call plot_data()\n";
	Fl::awake(plot_data);
exit_read:
	write_command("OFF\n");
}

bool get_swr(double f, double &swr)
{
	string data, linebuff;
	int numread = 0;
	int nread = 0;
	int numtries = 0;
	size_t p = string::npos;
	swr = 0;

	char szcmd[50];
	write_command("ON\n");

	f *= 1e6;
	f *= (1 + flaa_status.freq_corr / 1e6);
	snprintf(szcmd, sizeof(szcmd), "FQ%d\n", (int)f );
	if (!write_command(szcmd)) return false;

	if (!write_command("SW0\n")) return false;

	strcpy(szcmd, "FRX1\n");

	aa_ser_port.WriteBuffer(szcmd, strlen(szcmd));

	status_txt.assign("Write: ").append(szcmd);
	Fl::awake(update_status);

	numtries = 0;
	data.clear();
	memset(serbuff, 0, BUFFSIZE);
	for (;;) {
		if (read_exit) return false;
		memset(serbuff, 0, BUFFSIZE);
		nread = aa_ser_port.ReadBuffer(serbuff, BUFFSIZE);
		if (nread) {
			linebuff.assign("Read: ").append(serbuff);
			p = linebuff.find("\x0d");
			if (p != string::npos) {
				linebuff.erase(p);
				{
					guard_lock status_lock(&status_mutex);
					status_txt = linebuff;
					status_txt.append("\n");
					Fl::awake(update_status);
				}
			}
			data.append(serbuff);
			numread += nread;
		}

		if (data.find("OK") != string::npos)  break;
		if (data.find("ERROR") != string::npos)  return false;

		MilliSleep(5);
		numtries++;
		if (numtries * 0.005 > 4.0) {
			guard_lock status_lock(&status_mutex);
			status_txt = "READ FAILURE\n";
			Fl::awake(update_status);
			return false;
		}
	}

	float B, F, R, X, Z0 = flaa_status.system_z;
	p = data.find("\n");
	if (p != string::npos) {
		linebuff = data.substr(0, p);
		sscanf(linebuff.c_str(), "%f,%f,%f", &F, &R, &X);
		R /= Z0;
		if (R == 0) swr = 0;
		else {
			X /= Z0;
			B = (X * X + 1) / R + R;
			swr = (B + sqrtf( B * B - 4)) / 2.0;
		}
	}
	write_command("OFF\n");
	return true;
}

void read_swr(int b)
{
	switch (b) {
		default:
		case 80:
			if (get_swr(flaa_status.freq_80, swr_on_80))
				Fl::awake(update_swr_80);
			break;
		case 60:
			if (get_swr(flaa_status.freq_60, swr_on_60))
				Fl::awake(update_swr_60);
			break;
		case 40:
			if (get_swr(flaa_status.freq_40, swr_on_40))
				Fl::awake(update_swr_40);
			break;
		case 30:
			if (get_swr(flaa_status.freq_30, swr_on_30))
				Fl::awake(update_swr_30);
			break;
		case 20:
			if (get_swr(flaa_status.freq_20, swr_on_20))
				Fl::awake(update_swr_20);
			break;
		case 15:
			if (get_swr(flaa_status.freq_15, swr_on_15))
				Fl::awake(update_swr_15);
			break;
		case 10:
			if (get_swr(flaa_status.freq_10, swr_on_10))
				Fl::awake(update_swr_10);
			break;
		case 6:
			if (get_swr(flaa_status.freq_6, swr_on_6))
				Fl::awake(update_swr_6);
			break;
		case 2:
			if (get_swr(flaa_status.freq_2, swr_on_2))
				Fl::awake(update_swr_2);
			break;
		case 440:
			if (get_swr(flaa_status.freq_440, swr_on_440))
				Fl::awake(update_swr_440);
			break;
	}
}

double cf = 0, cr = 0, cx = 0;

void update_cable_values(void*)
{
	if (cf == 0 ) return;
	double c = 299.792458;
	if (!flaa_status.units)
		c = 983.571056;
	double len = (c * flaa_status.velocity_factor / cf) * flaa_status.wave;
	flaa_status.cable_freq = cf;
	flaa_status.cable_length = len;
	inp_cable_length->value(len);
	inp_cable_freq->value(cf);
}

void read_cable_values()
{
	int f1 = flaa_status.cable_freq * 1e6;
	int f2 = f1;

	double r1, x1, r2, x2, b, m;

	string data, linebuff;

	write_command("ON\n");
	write_command("SW0\n");

	for (int incr = 10000; incr > 1000; incr -= 1000) {
		f1 -= incr/2;
		f2 = f1 + incr;
		if (!get_rx(f1, r1, x1)) return;
		if (!get_rx(f2, r2, x2)) return;
//std::cout << "incr = " << incr << std::endl;
//std::cout << "f1 = " << f1 << ", r1 = " << r1 << ", x1 = " << x1 << std::endl;
//std::cout << "f2 = " << f2 << ", r2 = " << r2 << ", x2 = " << x2 << std::endl;
		if ( fabs(x1) < 0.01) {
			f2 = f1; x2 = x1; break;
		}
		if (fabs(x2) < 0.01) {
			f1 = f2; x1 = x2; break;
		}
		if ( fabs((x1+x2)/2) < 0.01) break;
		if (x1 == x2) break;
		m = (x2 - x1) / incr;
		b = x1 - (m * f1);
		f1 = -b/m;
	}
//std::cout << "f = " << (f1 + f2)/2 << ", X = " << (x2 + x1)/2.0 << std::endl;
	write_command("OFF\n");
	cf = ((f1 + f2)/2)*1e-6;
	cr = (r1 + r2)/2;
	cx = (x1 + x2)/2;
	Fl::awake(update_cable_values);
}

bool get_rx(int F, double &R, double &X){
	char szcmd[128];
	string data;
	char buffer[513];
	int nread = 0;
	size_t p = string::npos;

	double f = F * (1 + flaa_status.freq_corr / 1e6);
	snprintf(szcmd, sizeof(szcmd), "FQ%d\n", (int)f );
	if (!write_command(szcmd)) return false;

	write_command("FRX1\n");

	data.clear();
	for (;;) {
		if (read_exit) return false;
		memset(buffer, 0, 513);
		nread = aa_ser_port.ReadBuffer(buffer, 512);
		if (nread) data.append(buffer);
		if (data.find("OK") != string::npos)  break;
		if (data.find("ERROR") != string::npos)  return false;
	}
	p = data.find("\n");
	double df;
	if (p != string::npos) {
		data.erase(p);
//std::cout << data << std::endl;
		sscanf(data.c_str(), "%lf,%lf,%lf", &df, &R, &X);
	}
	return true;
}
